use crate::{
    command_handler::debrace, echo_to_term_frame, id_intern::intern_id, GlobalLayoutCell,
    GlobalLayoutState, GlobalMemoCell, TermFrame,
};
use gc_arena::{Gc, Rootable};
use piccolo::{
    self, async_sequence, meta_ops, Callback, CallbackReturn, Context, FromValue, Function,
    IntoValue, SequenceReturn, Table, UserData, Value, Variadic,
};
use std::{rc::Rc, str};
use yew::UseStateSetter;

use super::prep_metaop_call;

pub fn echo_frame_raw<'gc, 'a>(
    ctx: Context<'gc>,
    global_memo: &'a GlobalMemoCell,
    _global_layout: &'a UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame: TermFrame = try_unwrap_frame(
            ctx,
            &stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("Missing frame"))?,
        )?;
        let message: piccolo::String = stack.from_front(ctx)?;
        let message_str = str::from_utf8(message.as_bytes())
            .map_err(|_| "Expected message to echo to be UTF-8.".into_value(ctx))?;
        echo_to_term_frame(&global_memo, &frame, message_str).map_err(|m| m.into_value(ctx))?;
        Ok(piccolo::CallbackReturn::Return)
    })
}

pub fn echo_frame<'gc>(
    ctx: Context<'gc>,
    _global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> = ctx.get_global("commands")?;
        let function: Function = commands.get(ctx, "echo_frame_raw")?;
        let frame_no: Value = stack
            .pop_front()
            .ok_or_else(|| anyhow::Error::msg("Missing frame number"))?;
        let all_parts: Vec<String> = stack
            .consume::<Variadic<Vec<Value>>>(ctx)?
            .into_iter()
            .map(|v| format!("{:?}", v))
            .collect();
        stack.push_front(frame_no);
        let message = ctx.intern((all_parts.join(" ") + "\r\n").as_bytes());
        stack.push_back(message.into());
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

pub fn echo<'gc>(
    ctx: Context<'gc>,
    _global_memo: &GlobalMemoCell,
    _global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let commands: Table<'gc> = ctx.get_global("commands")?;
        let function: Function = commands.get(ctx, "echo_frame")?;
        let info: Table<'gc> = ctx.get_global("info")?;
        let cur_frame: Value = info.get(ctx, ctx.intern_static(b"current_frame"))?;
        stack.push_front(cur_frame);
        Ok(piccolo::CallbackReturn::Call {
            function,
            then: None,
        })
    })
}

pub fn vsplit<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let path: &str = debrace(&path);
        let frame: TermFrame = TermFrame(stack.from_front(ctx)?);
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .vsplit(path, frame.clone())
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(CallbackReturn::Call {
            function: Function::Callback(ensure_frame_instance(ctx, &frame)),
            then: None,
        })
    })
}

pub fn hsplit<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let path: &str = debrace(&path);
        let frame: TermFrame = TermFrame(stack.from_front(ctx)?);
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .hsplit(path, frame.clone())
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(CallbackReturn::Call {
            function: Function::Callback(ensure_frame_instance(ctx, &frame)),
            then: None,
        })
    })
}

pub fn panel_merge<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_layout = global_layout.clone();
    let global_memo = global_memo.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let path: String = stack.from_front(ctx)?;
        let path: &str = debrace(&path);
        let new_splits = global_memo
            .layout
            .borrow()
            .term_splits
            .join(path)
            .map_err(|e| e.into_value(ctx))?;
        let new_layout = Rc::new(GlobalLayoutState {
            term_splits: new_splits.clone(),
        });
        global_layout.set(new_layout.clone());
        *(global_memo.layout.borrow_mut()) = new_layout;
        Ok(piccolo::CallbackReturn::Return)
    })
}

pub fn try_unwrap_frame<'gc>(
    ctx: Context<'gc>,
    value: &Value<'gc>,
) -> Result<TermFrame, piccolo::Error<'gc>> {
    match u64::from_value(ctx, *value) {
        Ok(v) => Ok(TermFrame(v)),
        Err(_) => Ok(UserData::from_value(ctx, *value)?
            .downcast::<Rootable!['gcb => Gc<'gcb, TermFrame>]>()?
            .as_ref()
            .clone()),
    }
}

pub fn ensure_frame_instance<'gc>(ctx: Context<'gc>, frame: &TermFrame) -> Callback<'gc> {
    let frame = frame.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, _stack| {
        let frames: Table = ctx.get_global("frames")?;
        let frame_value: Value = frames.get(ctx, frame.0 as i64)?;
        if !frame_value.is_nil() {
            return Ok(CallbackReturn::Return);
        }

        let frame_tab = Table::new(&ctx);
        let classes: Table = ctx.get_global("classes")?;
        frame_tab.set_metatable(&ctx, Some(classes.get(ctx, "frame")?));
        frames.set(ctx, frame.0 as i64, frame_tab)?;

        // Call frame_tab:new(frame) to setup.
        let frame = frame.clone();
        let seq = async_sequence(&ctx, move |locals, mut seq| {
            let frame_tab = locals.stash(&ctx, frame_tab);
            async move {
                let call = seq.try_enter(|ctx, locals, _execution, mut stack| {
                    let frame_tab = locals.fetch(&frame_tab);
                    stack.consume(ctx)?;
                    Ok(prep_metaop_call(
                        ctx,
                        stack,
                        locals,
                        meta_ops::index(
                            ctx,
                            frame_tab.into_value(ctx),
                            ctx.intern_static(b"new").into_value(ctx),
                        )?,
                    ))
                })?;
                if let Some(call) = call {
                    seq.call(&call, 0).await?;
                }
                let call = seq.try_enter(|ctx, locals, _, mut stack| {
                    let value = stack
                        .pop_back()
                        .ok_or_else(|| anyhow::Error::msg("Index didn't return value"))?;
                    stack.consume(ctx)?;
                    stack.push_back(locals.fetch(&frame_tab).into_value(ctx));
                    stack.push_back(intern_id::<TermFrame>(ctx, frame.clone()));
                    Ok(locals.stash(&ctx, Function::from_value(ctx, value)?))
                })?;
                seq.call(&call, 0).await?;

                Ok(SequenceReturn::Return)
            }
        });
        Ok(CallbackReturn::Sequence(seq))
    })
}

pub(super) fn new_frame<'gc>(ctx: Context<'gc>, _global_memo: &GlobalMemoCell) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame_tab: Table = Table::from_value(
            ctx,
            stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("classes.frame:new missing object!"))?,
        )?;
        let frame: Value = stack
            .pop_front()
            .ok_or_else(|| anyhow::Error::msg("classes.frame:new missing frame!"))?;

        frame_tab.set(ctx, ctx.intern_static(b"frame"), frame)?;

        Ok(piccolo::CallbackReturn::Return)
    })
}

// Note: send_command_to_frame is called from the command parser, and can't be overriden (or even
// called) by the user.
// It will normally result in a call into frame_input, but that can be overriden by scripts.
pub(super) fn send_command_to_frame<'gc>(ctx: Context<'gc>) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let (frame, line): (Table, piccolo::String) = stack.consume(ctx)?;
        let seq = async_sequence(&ctx, |locals, mut seq| {
            let frame = locals.stash(&ctx, frame);
            let line = locals.stash(&ctx, line);
            async move {
                let call = seq.try_enter(|ctx, locals, _execution, stack| {
                    let frame = locals.fetch(&frame);
                    Ok(prep_metaop_call(
                        ctx,
                        stack,
                        locals,
                        meta_ops::index(
                            ctx,
                            frame.into_value(ctx),
                            ctx.intern_static(b"input").into_value(ctx),
                        )?,
                    ))
                })?;
                if let Some(call) = call {
                    seq.call(&call, 0).await?;
                }
                let call = seq.try_enter(|ctx, locals, _, mut stack| {
                    let value = stack
                        .pop_back()
                        .ok_or_else(|| anyhow::Error::msg("Index didn't return value"))?;
                    stack.consume(ctx)?;
                    stack.push_back(locals.fetch(&frame).into_value(ctx));
                    stack.push_back(locals.fetch(&line).into_value(ctx));
                    Ok(locals.stash(&ctx, Function::from_value(ctx, value)?))
                })?;
                seq.call(&call, 0).await?;

                Ok(SequenceReturn::Return)
            }
        });
        Ok(CallbackReturn::Sequence(seq))
    })
}

pub(super) fn frame_input<'gc>(ctx: Context<'gc>, _global_memo: &GlobalMemoCell) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let frame_tab: Table = Table::from_value(
            ctx,
            stack
                .pop_front()
                .ok_or_else(|| anyhow::Error::msg("classes.frame:new missing object!"))?,
        )?;
        let line: Value = stack
            .pop_front()
            .ok_or_else(|| anyhow::Error::msg("classes.frame:new missing line!"))?;

        let linked_mud = frame_tab.get_value(ctx, ctx.intern_static(b"linked_mud"));
        if linked_mud.is_nil() {
            return Ok(piccolo::CallbackReturn::Return);
        }
        let linked_mud: Table = Table::from_value(ctx, linked_mud)?;

        // linked_mud:mudinput_line(line)
        let seq = async_sequence(&ctx, |locals, mut seq| {
            let linked_mud = locals.stash(&ctx, linked_mud);
            let line = locals.stash(&ctx, line);
            async move {
                let call = seq.try_enter(|ctx, locals, _execution, mut stack| {
                    let linked_mud = locals.fetch(&linked_mud);
                    stack.consume(ctx)?;
                    Ok(prep_metaop_call(
                        ctx,
                        stack,
                        locals,
                        meta_ops::index(
                            ctx,
                            linked_mud.into_value(ctx),
                            ctx.intern_static(b"mudinput_line").into_value(ctx),
                        )?,
                    ))
                })?;
                if let Some(call) = call {
                    seq.call(&call, 0).await?;
                }
                let call = seq.try_enter(|ctx, locals, _, mut stack| {
                    let value = stack
                        .pop_back()
                        .ok_or_else(|| anyhow::Error::msg("Index didn't return value"))?;
                    stack.consume(ctx)?;
                    stack.push_back(locals.fetch(&linked_mud).into_value(ctx));
                    stack.push_back(locals.fetch(&line));
                    Ok(locals.stash(&ctx, Function::from_value(ctx, value)?))
                })?;
                seq.call(&call, 0).await?;

                Ok(SequenceReturn::Return)
            }
        });

        Ok(piccolo::CallbackReturn::Sequence(seq))
    })
}
